import os
import copy
import cv2
import numpy as np
import torch
from imageio import imread

from .models import pix2pix
from .detector import Detector

class Reconstructor:
    def __init__(self, weights_path=None, detector=None):
        if detector is None:
            detector = Detector()
        self.detector = detector
        self.unet = pix2pix.UNet()
        self.set_initial_weights(weights_path)
        self.unet.train()  # As in the original pix2pix, works as InstanceNormalization
    
    def set_initial_weights(self, weights_path):
        if weights_path is None:
            weights_path = os.path.join(os.path.dirname(__file__),
                    '../weights/faces_hybrid_and_rotated_2.pth')
            print('loading default reconstructor weights from {}'.format(weights_path))
        if not os.path.exists(weights_path):
            from .utils import download_from_gdrive
            from .constants import p2v_model_gdrive_id
            os.makedirs(os.path.dirname(weights_path), exist_ok=True)
            print('\tDownloading weights...')
            download_from_gdrive(p2v_model_gdrive_id, weights_path)
            print('\tDone!')
        self.initial_weights = torch.load(weights_path)

    def run(self, image, verbose=False):
        if type(image) is str:
            image = imread(image)
        image_cropped = self.detector.detect_and_crop(image)
        net_res = self.run_net(image_cropped)
        final_res = self.post_process(net_res)
        if verbose:
            from . import vis_depth_interactive
            vis_depth_interactive(final_res['Z_surface'])
        return final_res, image_cropped
        
    def run_net(self, img):
        # Because is actually instance normalization need to copy weights each time
        self.unet.load_state_dict(copy.deepcopy(self.initial_weights), strict=True)

        # Forward
        input = torch.from_numpy(img.transpose()).float()
        input = input.unsqueeze(0)
        input = input.transpose(2, 3)
        input = input.div(255.0).mul(2).add(-1)
        output = self.unet(input)
        output = output.add(1).div(2).mul(255)

        # Post Processing
        im_both = output.squeeze(0).detach().numpy().transpose().swapaxes(0, 1).copy()
        im_pncc = im_both[:, :, 0:3]
        im_depth = im_both[:, :, 3:6]
        im_depth[np.logical_and(im_depth < 10, im_depth > -10)] = 0
        im_pncc[np.logical_and(im_pncc < 10, im_pncc > -10)] = 0

        return {'pnnc': im_pncc, 'depth': im_depth}

    def post_process(self, net_res):
        im_pncc = net_res['pnnc'].astype(np.float64)
        im_depth = net_res['depth'].astype(np.float64)
        net_X = im_depth[:, :, 0] * (1.3674) / 255 - 0.6852
        net_Y = im_depth[:, :, 1] * (1.8401) / 255 - 0.9035
        net_Z = im_depth[:, :, 2] * (0.7542) / 255 - 0.2997
        mask = np.any(im_depth, axis=2) * np.all(im_pncc, axis=2)

        X = np.tile(np.linspace(-1, 1, im_depth.shape[1]), (im_depth.shape[0], 1))
        Y = np.tile(np.linspace(1, -1, im_depth.shape[0]).reshape(-1, 1), (1, im_depth.shape[1]))

        # Normalize fixed grid according to the network result, as X,Y are actually redundant
        X = (X - np.mean(X[mask])) / np.std(X[mask]) * np.std(net_X[mask]) + np.mean(net_X[mask])
        Y = (Y - np.mean(Y[mask])) / np.std(Y[mask]) * np.std(net_Y[mask]) + np.mean(net_Y[mask])

        Z = net_Z * 2  # Due to image resizing

        f = 1 / (X[0, 1] - X[0, 0])

        Z_surface = Z * f
        Z_surface[mask == False] = np.nan
        Z[mask == False] = np.nan

        return {'Z': Z, 'X': X, 'Y': Y, 'Z_surface': Z_surface}



from .reconstructor import Reconstructor
from .detector import Detector
from .utils import vis_net_result, vis_depth_interactive, vis_pcloud_interactive, vis_depth_matplotlib

reconstructor = None
def reconstruct(image=None, verbose=False):
    global reconstructor
    if reconstructor is None:
        reconstructor = Reconstructor()
    if image is None:
        import os
        from .constants import sample_image
        image = os.path.join(os.path.dirname(__file__), sample_image)
        print('No image specified, using {} as default input image'.format(image))
    return reconstructor.run(image, verbose)



import os
import dlib
import numpy as np
import math
import cv2
from .utils import download_url, extract_file


class Detector:
    def __init__(self, predictor_path=None):
        self.detector = dlib.get_frontal_face_detector()
        self.set_predictor(predictor_path)
    
    def set_predictor(self, predictor_path):
        if predictor_path is None:
            from .constants import predictor_file
            predictor_path = os.path.join(os.path.dirname(__file__), predictor_file)
            print('Loading default detector weights from {}'.format(predictor_path))
        if not os.path.exists(predictor_path):
            from .constants import predictor_url
            os.makedirs(os.path.dirname(predictor_path), exist_ok=True)
            print('\tDownloading weights from {}...'.format(predictor_url))
            download_url(predictor_url, save_path=os.path.dirname(predictor_path))
            print('\tExtracting weights...')
            extract_file(predictor_path + '.bz2', os.path.dirname(predictor_path))
            print('\tDone!')
        self.predictor = dlib.shape_predictor(predictor_path)

    def detect_and_crop(self, img, img_size=512):
        dets = self.detector(img, 1)  # Take a single detection
        for k, d in enumerate(dets):
            print("Detection {}: Left: {} Top: {} Right: {} Bottom: {}".format(
                k, d.left(), d.top(), d.right(), d.bottom()))
        dets = dets[0]
        shape = self.predictor(img, dets)
        points = shape.parts()

        pts = np.array([[p.x, p.y] for p in points])
        min_x = np.min(pts[:, 0])
        min_y = np.min(pts[:, 1])
        max_x = np.max(pts[:, 0])
        max_y = np.max(pts[:, 1])
        box_width = (max_x - min_x) * 1.2
        box_height = (max_y - min_y) * 1.2
        bbox = np.array([min_y - box_height * 0.3, min_x, box_height, box_width]).astype(np.int)

        img_crop = Detector.adjust_box_and_crop(img, bbox, crop_percent=150, img_size=img_size)
        # img_crop = img[bbox[0]:bbox[0]+bbox[2], bbox[1]:bbox[1]+bbox[3], :]
        return img_crop

    @staticmethod
    def adjust_box_and_crop(img, bbox, crop_percent=100, img_size=None):
        w_ext = math.floor(bbox[2])
        h_ext = math.floor(bbox[3])
        bbox_center = np.round(np.array([bbox[0] + 0.5 * bbox[2], bbox[1] + 0.5 * bbox[3]]))
        max_ext = np.round(crop_percent / 100 * max(w_ext, h_ext) / 2)
        top = max(1, bbox_center[0] - max_ext)
        left = max(1, bbox_center[1] - max_ext)
        bottom = min(img.shape[0], bbox_center[0] + max_ext)
        right = min(img.shape[1], bbox_center[1] + max_ext)
        height = bottom - top
        width = right - left
        # make the frame as square as possible
        if height < width:
            diff = width - height
            top_pad = int(max(0, np.floor(diff / 2) - top + 1))
            top = max(1, top - np.floor(diff / 2))
            bottom_pad = int(max(0, bottom + np.ceil(diff / 2) - img.shape[0]))
            bottom = min(img.shape[0], bottom + np.ceil(diff / 2))
            left_pad = 0
            right_pad = 0
        else:
            diff = height - width
            left_pad = int(max(0, np.floor(diff / 2) - left + 1))
            left = max(1, left - np.floor(diff / 2))
            right_pad = int(max(0, right + np.ceil(diff / 2) - img.shape[1]))
            right = min(img.shape[1], right + np.ceil(diff / 2))
            top_pad = 0
            bottom_pad = 0

        # crop the image
        img_crop = img[int(top):int(bottom), int(left):int(right), :]
        # pad the image
        img_crop = np.pad(img_crop, ((top_pad, bottom_pad), (left_pad, right_pad), (0, 0)), 'constant')

        if img_size is not None:
            img_crop = cv2.resize(img_crop, (img_size, img_size))
        return img_crop



predictor_url = 'http://dlib.net/files/shape_predictor_68_face_landmarks.dat.bz2'
predictor_file = '../weights/shape_predictor_68_face_landmarks.dat'
p2v_model_gdrive_id = '1op5_zyH4CWm_JFDdCUPZM4X-A045ETex'
sample_image = '../examples/sample.jpg'



import os
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm


def vis_depth_matplotlib(img, Z, elevation=60, azimuth=45, stride=5):
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm
    import matplotlib.pyplot as plt
    from matplotlib.colors import LightSource

    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # Create X and Y data
    x = np.arange(0, 512, 1)
    y = np.arange(0, 512, 1)
    X, Y = np.meshgrid(x, y)

    ls = LightSource(azdeg=0, altdeg=90)
    # shade data, creating an rgb array.
    img_c = np.concatenate((img.astype('float') / 255, np.ones((512, 512, 1))), axis=2)
    rgb = ls.shade_rgb(img.astype('float') / 255, Z, blend_mode='overlay')

    surf = ax.plot_surface(X, Y, Z, cstride=stride, rstride=stride, linewidth=0, antialiased=False, facecolors=rgb)
    ax.view_init(elev=elevation, azim=azimuth)
    # ax.view_init(elev=90., azim=90)`
    # Show the plot
    plt.show()


def vis_depth_interactive(Z):
    import k3d
    Nx, Ny = 1, 1
    xmin, xmax = 0, Z.shape[0]
    ymin, ymax = 0, Z.shape[1]

    x = np.linspace(xmin, xmax, Nx)
    y = np.linspace(ymin, ymax, Ny)
    x, y = np.meshgrid(x, y)
    plot = k3d.plot(grid_auto_fit=True, camera_auto_fit=False)
    plt_surface = k3d.surface(-Z.astype(np.float32), color=0xb2ccff,
                              bounds=[xmin, xmax, ymin, ymax])  # -Z for mirroring
    plot += plt_surface
    plot.display()
    plot.camera = [242.57934019166004, 267.50948550191197, -406.62328311352337, 256, 256, -8.300323486328125,
                   -0.13796270478729053, -0.987256298362836, -0.07931767413815752]
    return plot


def vis_pcloud_interactive(res, img):
    import k3d
    from colormap import rgb2hex
    color_vals = np.zeros((img.shape[0], img.shape[1]))
    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            color_vals[i, j] = int(rgb2hex(img[i, j, 0], img[i, j, 1], img[i, j, 2]).replace('#', '0x'), 0)
    colors = color_vals.flatten()

    points = np.stack((res['X'].flatten(), res['Y'].flatten(), res['Z'].flatten()), axis=1)

    invalid_inds = np.any(np.isnan(points), axis=1)
    points_valid = points[invalid_inds == False]
    colors_valid = colors[invalid_inds == False]
    plot = k3d.plot(grid_auto_fit=True, camera_auto_fit=False)
    plot += k3d.points(points_valid, colors_valid, point_size=0.01, compression_level=9, shader='flat')
    plot.display()
    plot.camera = [-0.3568942548181382, -0.12775125650240726, 3.5390732533009452, 0.33508163690567017,
                   0.3904658555984497, -0.0499117374420166, 0.11033077266672488, 0.9696364582197756, 0.2182481603445357]
    return plot


def vis_net_result(img, net_result):
    plt.figure()
    plt.subplot(1, 3, 1)
    plt.imshow(img)
    plt.title('Input Image')
    plt.subplot(1, 3, 2)
    plt.imshow(net_result['pnnc'].astype(np.uint8))
    plt.title('PNCC Visualization')
    plt.subplot(1, 3, 3)
    plt.imshow(net_result['depth'][:, :, 2].astype(np.uint8), cmap='gray')
    plt.title('Depth Visualization')
    plt.show()


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def download_url(url, save_path):
    from six.moves import urllib
    save_path = os.path.expanduser(save_path)
    if not os.path.exists(save_path):
        makedir(save_path)

    filename = url.rpartition('/')[2]
    filepath = os.path.join(save_path, filename)

    try:
        with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1,
                      desc=url.split('/')[-1]) as t:  # all optional kwargs
            urllib.request.urlretrieve(url, filepath, reporthook=t.update_to)
            t.total = t.n
    except ValueError:
        raise Exception('Failed to download! Check URL: ' + url +
                        ' and local path: ' + save_path)


def extract_file(path, to_directory=None):
    path = os.path.expanduser(path)
    if path.endswith('.zip'):
        opener, mode = zipfile.ZipFile, 'r'
    elif path.endswith(('.tar.gz', '.tgz')):
        opener, mode = tarfile.open, 'r:gz'
    elif path.endswith(('tar.bz2', '.tbz')):
        opener, mode = tarfile.open, 'r:bz2'
    elif path.endswith('.bz2'):
        import bz2
        opener, mode = bz2.BZ2File, 'rb'
        with open(path[:-4], 'wb') as fp_out, opener(path, 'rb') as fp_in:
            for data in iter(lambda: fp_in.read(100 * 1024), b''):
                fp_out.write(data)
        return
    else:
        raise (ValueError,
               "Could not extract `{}` as no extractor is found!".format(path))

    if to_directory is None:
        to_directory = os.path.abspath(os.path.join(path, os.path.pardir))
    cwd = os.getcwd()
    os.chdir(to_directory)

    try:
        file = opener(path, mode)
        try:
            file.extractall()
        finally:
            file.close()
    finally:
        os.chdir(cwd)


def download_from_gdrive(id, destination):
    import requests
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params={'id': id}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id': id, 'confirm': token}
        response = session.get(URL, params=params, stream=True)

    save_response_content(response, destination)


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


def save_response_content(response, destination):
    CHUNK_SIZE = 32768
    t = tqdm(unit='B', unit_scale=True, miniters=1)

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                t.update(len(chunk))
                f.write(chunk)



import torch
import torch.nn as nn

def conv_block(in_channels, out_channels):
    return nn.Sequential(
        nn.LeakyReLU(0.2,inplace=True),
        nn.Conv2d(in_channels, out_channels, 4, stride=2,padding=1),
        nn.BatchNorm2d(out_channels)
    )


def deconv_block(in_channels, out_channels,use_dropout=False):
    layers = [
        nn.ReLU(inplace=True),
        nn.ConvTranspose2d(in_channels, out_channels, 4, stride=2,padding=1),
        nn.BatchNorm2d(out_channels)
    ]
    if use_dropout:
        layers.append(nn.Dropout(0.5))
    return nn.Sequential(*layers)

class UNet(nn.Module):

    def __init__(self):
        super().__init__()

        self.conv_down1 = nn.Conv2d(3,64,4,stride=2,padding=1)

        self.conv_down2 = conv_block(64,128)

        self.conv_down3 = conv_block(128,256)

        self.conv_down4 = conv_block(256,512)

        self.conv_down5 = conv_block(512,512)

        self.conv_down6 = conv_block(512,512)

        self.conv_down7 = conv_block(512,512)

        self.conv_down8 = conv_block(512,512)

        self.conv_up1 = deconv_block(512,512,use_dropout=True)

        self.conv_up2 = deconv_block(1024,512,use_dropout=True)

        self.conv_up3 = deconv_block(1024,512,use_dropout=True)

        self.conv_up4 = deconv_block(1024,512)

        self.conv_up5 = deconv_block(1024,256)

        self.conv_up6 = deconv_block(512,128)

        self.conv_up7 = deconv_block(256,64)

        self.conv_up8 = deconv_block(128,64)

        self.conv_up9 = nn.Sequential(
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(64, 64, 3, stride=1,padding=1),
            nn.BatchNorm2d(64)
        )

        self.conv_up9 = nn.Sequential(
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(64, 64, 3, stride=1,padding=1),
            nn.BatchNorm2d(64)
        )

        self.conv_up10 = nn.Sequential(
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(64, 32, 3, stride=1,padding=1),
            nn.BatchNorm2d(32)
        )

        self.conv_up11 = nn.Sequential(
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(32, 7, 3, stride=1,padding=1)
        )

    # TODO: rewrite nicely

    def forward(self, x):
        down1 = self.conv_down1(x)
        down2 = self.conv_down2(down1)
        down3 = self.conv_down3(down2)
        down4 = self.conv_down4(down3)
        down5 = self.conv_down5(down4)
        down6 = self.conv_down6(down5)
        down7 = self.conv_down7(down6)
        down8 = self.conv_down8(down7)

        up1 = self.conv_up1(down8)
        up2 = self.conv_up2(torch.cat([up1, down7], dim=1))
        up3 = self.conv_up3(torch.cat([up2, down6], dim=1))
        up4 = self.conv_up4(torch.cat([up3, down5], dim=1))
        up5 = self.conv_up5(torch.cat([up4, down4], dim=1))
        up6 = self.conv_up6(torch.cat([up5, down3], dim=1))
        up7 = self.conv_up7(torch.cat([up6, down2], dim=1))
        up8 = self.conv_up8(torch.cat([up7, down1], dim=1))
        up9 = self.conv_up9(up8)
        up10 = self.conv_up10(up9)
        up11 = self.conv_up11(up10)

        return up11



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# MIT License
#   
# Copyright (c) 2020 Elad Richardson and Matan Sela
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import setuptools

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as fh:
    long_description = fh.read()

with open(os.path.join(os.path.dirname(__file__), "requirements.txt")) as f:
    required = f.read().splitlines()

setuptools.setup(
    name="pix2vertex", # Replace with your own username
    version="1.0.3",
    author="Elad Richardson, Matan Sela",
    author_email="elad.richardson@gmail.com, matansel@gmail.com",
    description="3D face reconstruction from a single image",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/eladrich/pix2vertex.pytorch",
    packages=setuptools.find_packages(exclude=["tests.*", "tests"]),
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=required,
    include_package_data=True,
    keywords="pix2vertex face reconstruction 3d pytorch pip package",
    python_requires='>=3.4',
    zip_safe=False
)
